# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore
import time
from typing import Dict, NamedTuple

import dm_env
import gym
import numpy as np
from acme import specs, types


def rad_rot_to_xy(rad_rot):
    return np.cos(rad_rot), np.sin(rad_rot)


def deg_rot_to_xy(deg_rot):
    return np.cos(deg_rot * np.pi / 180), np.sin(deg_rot * np.pi / 180)


def should_wait(wait_list):
    should_wait = False
    for entity in wait_list:
        if not entity.wm.new_data:
            should_wait = True
            break
    return should_wait


def wait_for_next_observations(obs_to_wait_for):
    """
    Wait for next observation before agents should think again.
    """
    # Wait for new data
    waiting = should_wait(obs_to_wait_for)

    # start = time.time()
    while waiting:
        waiting = should_wait(obs_to_wait_for)
        time.sleep(0.0001)
    # end = time.time()
    # print("Wait time: ", end-start)

    # Set new data false
    for entity in obs_to_wait_for:
        entity.wm.new_data = False


# Dummy class to get the observation in the correct format
class OLT(NamedTuple):
    """Container for (observation, legal_actions, terminal) tuples."""

    observation: types.Nest
    legal_actions: types.Nest
    terminal: types.Nest


class SpecWrapper(dm_env.Environment):
    """Spec wrapper for 2D RoboCup environment."""

    # Note: we don't inherit from base.EnvironmentWrapper because that class
    # assumes that the wrapped environment is a dm_env.Environment.

    def __init__(self, num_players: int):
        self._reset_next_step = True

        self.scaling = 200.0

        # Chose action
        act_min = [0.0] * 7  # 6 + No action
        act_max = [1.0] * 7  # 6 + No action

        # Action continuous component
        # All directions are in x, y format
        act_min.extend(
            [
                -100 / self.scaling,
                -1,
                -1,  # dash (power, direction)
                0,
                -1,
                -1,  # kick (power, direction)
                0,
                0,  # change_view (width, quality)
                -1,
                -1,
                0,  # tackle (direction, foul)
                -1,
                -1,  # turn (direction)
                -1,
                -1,
            ]
        )  # turn_neck(direction)

        act_max.extend([100 / self.scaling, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

        assert len(act_min) == len(act_max)
        action_spec = specs.BoundedArray(
            shape=(len(act_min),),
            dtype="float32",
            name="action",
            minimum=act_min,
            maximum=act_max,
        )

        self.action_size = action_spec.shape[0]

        # obs_dict = {"time_left": 0, "side": 1, "sense_self": 2,
        #  "coords": (3, 5), "body_dir": (5, 7),
        #             "head_dir": (7, 9), "width": (9, 12),
        # "quality": 13, "stamina": 14, "effort": 15,
        #             "speed_amount": 16, "speed_direction": (17, 19),
        # "neck_direction": (19, 21),
        #             "see_ball": 21, "ball_dist": 22,
        # "ball_dir": (23, 25), "ball_dist_change": 25,
        #             "ball_dir_change": 26, "ball_speed": 27,
        # "last_action": (28, 28 + self.action_size),
        #             }

        # TODO: Check if all bounds are correct
        obs_min = [
            0.0,  # time_left
            0.0,  # side
            0.0,  # sense_self
            -100 / self.scaling,
            -50 / self.scaling,  # coords
            -1,
            -1,  # body_dir
            -1,
            -1,  # head_dir
            0,
            0,
            0,  # width
            0,  # quality
            0,  # stamina
            0,  # effort
            0,  # speed_amount
            -1,
            -1,  # speed_direction
            -1,
            -1,  # neck_direction
            0,  # see_ball
            0,  # ball_dist
            -1,
            -1,  # ball_dir
            -100 / self.scaling,  # ball_dist_change
            -180 / self.scaling,  # ball_dir_change
            0,  # ball_speed
        ]

        obs_max = [
            1.0,  # time_left
            1.0,  # side
            1.0,  # sense_self
            100 / self.scaling,
            50 / self.scaling,  # coords
            1,
            1,  # body_dir
            1,
            1,  # head_dir
            1,
            1,
            1,  # width
            1,  # quality
            1,  # stamina
            1,  # effort
            100 / self.scaling,  # speed_amount
            1,
            1,  # speed_direction
            1,
            1,  # neck_direction
            1,  # see_ball
            100 / self.scaling,  # ball_dist
            1,
            1,  # ball_dir
            100 / self.scaling,  # ball_dist_change
            180 / self.scaling,  # ball_dir_change
            100 / self.scaling,  # ball_speed
        ]

        # Last action
        obs_min.extend(action_spec.minimum)
        obs_max.extend(action_spec.maximum)

        # [see_player, is_on_team, player_distance,
        # player_direction] for num_agents-1
        self.num_agents = num_players

        # TODO: Add this in again.
        # for i in range(21):
        #     # [see_player, is_on_team, player_distance,
        # player_direction (x, y format)]
        #     obs_min.extend([0, 0, -200 / self.scaling, -1, -1])
        #     obs_max.extend([1, 1, +200 / self.scaling, 1, 1])

        assert len(obs_min) == len(obs_max)
        self.obs_size = len(obs_min)

        self.agents = ["player_" + str(r) for r in range(num_players)]

        self._observation_specs = {}
        self._action_specs = {}

        obs_spec = specs.BoundedArray(
            shape=(self.obs_size,),
            dtype="float32",
            name="observation",
            minimum=obs_min,
            maximum=obs_max,
        )

        # Time_left, ball coords, ball delta_coords
        state_min = [0, -100 / self.scaling, -100 / self.scaling, -10, -10]
        state_max = [1, 100 / self.scaling, 100 / self.scaling, 10, 10]

        # First player is the critic player
        # Players sides,  coords, delta_coords, body_angle (x, y format),
        # head_angle (x, y format)
        for i in range(num_players):
            state_min.extend(
                [
                    0.0,
                    -100 / self.scaling,
                    -100 / self.scaling,
                    -10,
                    -10,
                    -1,
                    -1,
                    -1,
                    -1,
                ]
            )
            state_max.extend(
                [1.0, +100 / self.scaling, +100 / self.scaling, +10, +10, 1, 1, 1, 1]
            )

        # Add all observations to state info
        for i in range(num_players):
            state_min.extend(obs_min)
            state_max.extend(obs_max)

        assert len(state_min) == len(state_max)
        self._state_spec = specs.BoundedArray(
            shape=(len(state_min),),
            dtype="float32",
            name="state",
            minimum=state_min,
            maximum=state_max,
        )

        self._discount = dict(zip(self.agents, [np.float32(1.0)] * len(self.agents)))

        # TODO: Delete this
        # self.previous_act = {"player_0": None}

        for agent in self.agents:
            # TODO: Why is the action spec in two places?
            self._observation_specs[agent] = OLT(
                observation=obs_spec,
                legal_actions=action_spec,
                terminal=specs.Array((1,), np.float32),
            )

            self._action_specs[agent] = action_spec
        # Obs: BoundedArray(shape=(4,), dtype=dtype('float32'), name='observation',
        # minimum=[-4.8000002e+00 -3.4028235e+38 -4.1887903e-01 -3.4028235e+38],
        #  maximum=[4.8000002e+00 3.4028235e+38 4.1887903e-01 3.4028235e+38])
        # Actions: DiscreteArray(shape=(), dtype=int64, name=action, minimum=0,
        #  maximum=1, num_values=2)

    def reset(self):
        pass

    def step(self):
        pass

    def observation_spec(self) -> types.NestedSpec:
        return self._observation_specs

    def action_spec(self) -> types.NestedSpec:
        return self._action_specs

    def reward_spec(self) -> Dict[str, specs.Array]:
        reward_specs = {}
        for agent in self.agents:
            reward_specs[agent] = specs.Array((), np.float32)
        return reward_specs

    def discount_spec(self) -> Dict[str, specs.BoundedArray]:
        discount_specs = {}
        for agent in self.agents:
            discount_specs[agent] = specs.BoundedArray(
                (), np.float32, minimum=0, maximum=1.0
            )
        return discount_specs

    def extra_spec(self) -> Dict[str, specs.BoundedArray]:
        return {"s_t": self._state_spec}

    def _proc_robocup_obs(
        self, observations: Dict, done: bool, nn_actions: Dict = None
    ) -> Dict:
        # TODO: Try to automatically normalise by min max boundries
        processed_obs_dict = {}
        for obs_i, agent_key in enumerate(self.agents):
            env_agent_obs = observations[agent_key]

            if nn_actions:
                last_action = nn_actions[agent_key]
            else:
                last_action = None

            proc_agent_obs = self.proc_agent_env_obs(env_agent_obs, last_action)

            observation = OLT(
                observation=proc_agent_obs,
                legal_actions=np.ones(self.action_size, dtype=np.float32),
                terminal=np.asarray([done], dtype=np.float32),
            )
            processed_obs_dict[agent_key] = observation
        return processed_obs_dict

    def proc_agent_env_obs(self, env_agent_obs, last_action):  # noqa: C901
        # All angles is in x,y format
        proc_agent_obs = np.zeros(self.obs_size, dtype=np.float32)
        obs_dict = {
            "time_left": 0,
            "side": 1,
            "sense_self": 2,
            "coords": (3, 5),
            "body_dir": (5, 7),
            "head_dir": (7, 9),
            "width": (9, 12),
            "quality": 12,
            "stamina": 13,
            "effort": 14,
            "speed_amount": 15,
            "speed_direction": (16, 18),
            "neck_direction": (18, 20),
            "see_ball": 20,
            "ball_dist": 21,
            "ball_dir": (22, 24),
            "ball_dist_change": 24,
            "ball_dir_change": 25,
            "ball_speed": 26,
            "last_action": (27, 27 + self.action_size),
        }

        # TODO: Fix this when there is other players
        assert self.obs_size == 27 + self.action_size

        # Time left obs
        if "game_step" in env_agent_obs and "game_length" in env_agent_obs:
            # Time left
            proc_agent_obs[obs_dict["time_left"]] = (
                1 - env_agent_obs["game_step"] / env_agent_obs["game_length"]
            )

        # Team side obs
        side_to_id = {"l": 0.0, "r": 1.0}
        proc_agent_obs[1] = side_to_id[env_agent_obs["side"]]

        if (
            "estimated_abs_coords" in env_agent_obs
            and env_agent_obs["estimated_abs_coords"][0] is not None
        ):
            # see_own_stats
            proc_agent_obs[obs_dict["sense_self"]] = 1.0

            # coords
            coords = env_agent_obs["estimated_abs_coords"]
            s, e = obs_dict["coords"]
            proc_agent_obs[s] = float(coords[0]) / self.scaling
            proc_agent_obs[e - 1] = float(coords[1]) / self.scaling

            # body_angle
            s, e = obs_dict["body_dir"]
            proc_agent_obs[s:e] = deg_rot_to_xy(env_agent_obs["estimated_abs_body_dir"])

            # head_angle
            s, e = obs_dict["head_dir"]
            proc_agent_obs[s:e] = deg_rot_to_xy(env_agent_obs["estimated_abs_neck_dir"])

            # view_width
            w_to_int = {"narrow": 0, "normal": 1, "wide": 2}
            onehot = np.zeros(3)
            onehot[w_to_int[env_agent_obs["view_width"]]] = 1
            s, e = obs_dict["width"]
            proc_agent_obs[s:e] = onehot

            # view_quality
            q_to_int = {"high": 0, "low": 1}
            proc_agent_obs[obs_dict["quality"]] = q_to_int[
                env_agent_obs["view_quality"]
            ]

            # stamina
            proc_agent_obs[obs_dict["stamina"]] = env_agent_obs["stamina"] / 8000

            # effort
            proc_agent_obs[obs_dict["effort"]] = env_agent_obs["effort"]

            # speed_amount
            proc_agent_obs[obs_dict["speed_amount"]] = (
                env_agent_obs["speed_amount"] / self.scaling
            )

            # speed_dir
            s, e = obs_dict["speed_direction"]
            proc_agent_obs[s:e] = deg_rot_to_xy(env_agent_obs["speed_direction"])

            # Relative neck dir
            s, e = obs_dict["neck_direction"]
            proc_agent_obs[s:e] = deg_rot_to_xy(env_agent_obs["neck_direction"])

        # See_ball, ball_distance, ball_direction, dist_change, dir_change, speed
        if "ball" in env_agent_obs and env_agent_obs["ball"] is not None:
            # Has ball flag
            proc_agent_obs[obs_dict["see_ball"]] = 1.0

            if env_agent_obs["ball"].distance is not None:
                proc_agent_obs[obs_dict["ball_dist"]] = (
                    env_agent_obs["ball"].distance / self.scaling
                )

            if env_agent_obs["ball"].direction is not None:
                s, e = obs_dict["ball_dir"]
                proc_agent_obs[s:e] = deg_rot_to_xy(env_agent_obs["ball"].direction)

            if env_agent_obs["ball"].dist_change is not None:
                proc_agent_obs[obs_dict["ball_dist_change"]] = (
                    env_agent_obs["ball"].dist_change / self.scaling
                )

            if env_agent_obs["ball"].dir_change is not None:
                proc_agent_obs[obs_dict["ball_dir_change"]] = (
                    env_agent_obs["ball"].dir_change / self.scaling
                )

            if env_agent_obs["ball"].speed is not None:
                proc_agent_obs[obs_dict["ball_speed"]] = (
                    env_agent_obs["ball"].speed / self.scaling
                )

        # Last player actions
        if last_action is not None:
            s, e = obs_dict["last_action"]
            proc_agent_obs[s:e] = last_action

        # [see_player, is_on_team, distance, direction, dist_change,
        # dir_change, speed, body_direction, neck_direction]
        if "players" in env_agent_obs and len(env_agent_obs["players"]) > 0:
            raise NotImplementedError(
                "Obs for more than one player not implemented yet."
            )
            # ["opponent", player.distance, player.direction (x, y format)]
            num_see_players = len(env_agent_obs["players"])

            # Place the players on the closest available spot
            players = env_agent_obs["players"]

            # TODO: Get better sorting algorithm. This one is good,
            # but also allows for player degrees not to
            #  always be assending. This is because a player can be
            # selected for a slot and then the player with a
            #  bigger positive degree might be selected for a lower
            # slot. Therefore this might happend:
            #  player[i].degree > player[i+1].degree.

            player_ids = list(range(num_see_players))
            spots = [[d_i, (d_i / 20) * 180 - 90] for d_i in range(21)]
            player_spots = [-1] * num_see_players
            for _ in range(num_see_players):
                best_p_i = None
                beset_spot = None
                best_s_count = None
                best_p_count = None
                beset_dist = np.inf

                # All remaining players select the closest available spot
                for p_count, p_i in enumerate(player_ids):
                    player_type, player_dist, player_dir = players[p_i]

                    # Go through all the spots and get the closest one to the player
                    for s_count, spot in enumerate(spots):
                        s_i, s_dir = spot

                        dist_to_spot = abs(s_dir - player_dir)

                        if dist_to_spot < beset_dist:
                            beset_dist = dist_to_spot
                            best_p_i = p_i
                            beset_spot = s_i
                            best_p_count = p_count
                            best_s_count = s_count

                # Give the closest player its spot.
                player_ids.pop(best_p_count)
                spots.pop(best_s_count)
                player_spots[best_p_i] = beset_spot

            assert len(player_ids) == 0
            # Calculate all the available slots to put player in
            slots = [
                [11 + self.action_size + s_i * 5, 11 + self.action_size + s_i * 5 + 5]
                for s_i in player_spots
            ]
            # slots = [i, i + 5] for i in
            # slots = [all_slots[p_spot] for p_spot in player_spots]

            assert len(slots) == num_see_players
            player_type_dict = {"opponent": 0.0, "team": 1.0}

            for p_i, player in enumerate(players):
                player_type, player_dist, player_dir = player

                # pop a random element from the slots list
                start_i, end_i = slots[p_i]

                # see_player, is_on_team, player_distance,
                # player_direction (x, y format)
                dir_x, dir_y = deg_rot_to_xy(player_dir)

                proc_agent_obs[start_i:end_i] = [
                    1.0,
                    player_type_dict[player_type],
                    player_dist / self.scaling,
                    dir_x,
                    dir_y,
                ]

        # if not env_agent_obs["obs_updated"]:
        #     proc_agent_obs[2] = 0.0

        return proc_agent_obs

    def _proc_robocup_state(self, state: Dict, proc_obs: Dict) -> np.array:
        state_dict = {
            "time_left": 0,
            "ball_coords": (1, 3),
            "ball_delta": (3, 5),
            "player_offset": 5,
            "p_obs_size": 9,
            "p_side": 0,
            "p_coords": (1, 3),
            "p_delta": (3, 5),
            "p_b_dir": (5, 7),
            "p_n_dir": (7, 9),
        }
        offset = state_dict["player_offset"]

        proc_agent_state = np.zeros(
            offset + self.num_agents * (9 + self.obs_size), dtype=np.float32
        )

        # Time left
        proc_agent_state[state_dict["time_left"]] = (
            1 - state["game_step"] / state["game_length"]
        )

        # Ball:
        ball = state["ball"]
        s, e = state_dict["ball_coords"]
        proc_agent_state[s:e] = [
            float(ball["coords"][0]) / self.scaling,
            float(ball["coords"][1]) / self.scaling,
        ]
        s, e = state_dict["ball_delta"]
        proc_agent_state[s:e] = [
            float(ball["delta_coords"][0]),
            float(ball["delta_coords"][1]),
        ]  # TODO: Should delta coords have a normaliser?

        # TODO: Add check to see if players are in the correct order
        # TODO: Include onehot indicating which player should be focussed on.
        # Also use dictionary of states. Players should not be reordered as this
        # messes with the action order provided to the Q value critic.
        players = state["players"]
        p_size = state_dict["p_obs_size"]
        if players:
            for i in range(self.num_agents):
                if len(players) > i:
                    player = players[i]
                    # 'side': 0, 'coords': (52.6498, 0.54963),
                    # 'delta_coords': (0.000227909, 0.00371977), 'body_angle': 163,
                    # 'neck_angle': 0}
                    proc_agent_state[
                        i * p_size + offset + state_dict["p_side"]
                    ] = player["side"]
                    s, e = state_dict["p_coords"]
                    proc_agent_state[
                        i * p_size + offset + s : i * p_size + offset + e
                    ] = [
                        player["coords"][0] / self.scaling,
                        player["coords"][1] / self.scaling,
                    ]
                    s, e = state_dict["p_delta"]
                    proc_agent_state[
                        i * p_size + offset + s : i * p_size + offset + e
                    ] = player[
                        "delta_coords"
                    ]  # TODO: Should delta coords have a normaliser?
                    s, e = state_dict["p_b_dir"]
                    proc_agent_state[
                        i * p_size + offset + s : i * p_size + offset + e
                    ] = deg_rot_to_xy(player["body_angle"])
                    s, e = state_dict["p_n_dir"]
                    proc_agent_state[
                        i * p_size + offset + s : i * p_size + offset + e
                    ] = deg_rot_to_xy(player["neck_angle"])

                else:
                    break

        # Add agent observations
        start_i = offset + self.num_agents * p_size
        for agent_i, agent_key in enumerate(self.agents):
            obs = proc_obs[agent_key].observation
            proc_agent_state[
                start_i
                + agent_i * self.obs_size : start_i
                + (agent_i + 1) * self.obs_size
            ] = obs

        return proc_agent_state

    def _proc_robocup_actions(self, actions: Dict) -> Dict:
        # dash (speed), turn (direction), kick (power, direction)
        processed_action_dict = {}

        for agent_key in actions.keys():
            action = actions[agent_key]
            assert len(action) == self.action_size
            processed_action_dict[agent_key] = self.proc_agent_action(action)

        return processed_action_dict

    def proc_agent_action(self, action):
        # TODO: Add catch, tackle, move, pointto, attentionto and say commands as well.
        int_to_command = [
            "dash",
            "kick",
            "change_view",
            "tackle",
            "turn",
            "turn_neck",
            "none",
        ]

        # Remove change_view and turn_neck action
        # TODONE: Remove this
        # action[2] = 0
        # action[5] = 0

        command = int_to_command[np.argmax(action[0 : len(int_to_command)])]

        # Do the command
        assert len(int_to_command) == 7

        act_dict = {
            "dash_pow": 7,
            "dash_dir": (8, 10),
            "kick_pow": 10,
            "kick_dir": (11, 13),
            "width": 13,
            "quality": 14,
            "tackle_dir": (15, 17),
            "tackle_foul": 17,
            "turn": (18, 20),
            "neck": (20, 22),
        }

        if command == "dash":
            power = action[act_dict["dash_pow"]] * self.scaling
            s, e = act_dict["dash_dir"]
            dir_x, dir_y = action[s:e]
            dir = np.arctan2(dir_y, dir_x) * 180 / np.pi
            robocup_action = "(dash " + str(power) + " " + str(dir) + ")"
        elif command == "kick":
            power = action[act_dict["kick_pow"]] * self.scaling
            s, e = act_dict["kick_dir"]
            dir_x, dir_y = action[s:e]
            dir = np.arctan2(dir_y, dir_x) * 180 / np.pi
            robocup_action = "(kick " + str(power) + " " + str(dir) + ")"
        elif command == "change_view":
            w_to_text = ["narrow", "normal", "wide"]
            width = w_to_text[int(action[act_dict["width"]] * 2.99)]
            q_to_text = ["high", "low"]
            quality = q_to_text[int(action[act_dict["quality"]] * 1.99)]
            robocup_action = "(change_view " + width + " " + quality + ")"
        elif command == "tackle":
            s, e = act_dict["tackle_dir"]
            dir_x, dir_y = action[s:e]
            dir = np.arctan2(dir_y, dir_x) * 180 / np.pi
            f_to_text = ["true", "false"]
            foul = f_to_text[int(action[act_dict["tackle_foul"]] * 1.99)]
            robocup_action = "(tackle " + str(dir) + " " + foul + ")"
        elif command == "turn":
            s, e = act_dict["turn"]
            dir_x, dir_y = action[s:e]
            dir = np.arctan2(dir_y, dir_x) * 180 / np.pi
            robocup_action = "(turn " + str(dir) + ")"
        elif command == "turn_neck":
            s, e = act_dict["neck"]
            x, y = action[s:e]
            turn_neck_dir = np.arctan2(y, x) * 180 / np.pi
            robocup_action = "(turn_neck " + str(turn_neck_dir) + ")"
        elif command == "none":
            robocup_action = "(done)"
        else:
            raise NotImplementedError("Command not implemented: ", command)
        return robocup_action

    @property
    def possible_agents(self) -> gym.Env:
        """Returns the number of possible agents."""
        return self.agents



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MARL system Builder interface."""

import abc
from typing import Dict, Iterator, List, Optional

import reverb
import sonnet as snt
from acme import specs
from acme.utils import counting

from mava import adders, core, types


class SystemBuilder(abc.ABC):
    """Defines an interface for defining the components of an RL system.
    Implementations of this interface contain a complete specification of a
    concrete RL system. An instance of this class can be used to build an
    RL system which interacts with the environment either locally or in a
    distributed setup.
    """

    @abc.abstractmethod
    def make_replay_tables(
        self,
        environment_spec: specs.EnvironmentSpec,
    ) -> List[reverb.Table]:
        """Create tables to insert data into."""

    @abc.abstractmethod
    def make_dataset_iterator(
        self,
        replay_client: reverb.Client,
    ) -> Iterator[reverb.ReplaySample]:
        """Create a dataset iterator to use for learning/updating the agent."""

    @abc.abstractmethod
    def make_adder(
        self,
        replay_client: reverb.Client,
    ) -> Optional[adders.ParallelAdder]:
        """Create an adder which records data generated by the executor/environment.
        Args:
          replay_client: Reverb Client which points to the replay server.
        """

    @abc.abstractmethod
    def make_executor(
        self,
        policy_networks: Dict[str, snt.Module],
        adder: Optional[adders.ParallelAdder] = None,
        variable_source: Optional[core.VariableSource] = None,
    ) -> core.Executor:
        """Create an executor instance.
        Args:
          policy_networks: A struct of instance of all the different
            policy networks; this should be a callable
            which takes as input observations and returns actions.
          adder: How data is recorded (e.g. added to replay).
          variable_source: A source providing the necessary actor parameters.
        """

    @abc.abstractmethod
    def make_trainer(
        self,
        networks: Dict[str, Dict[str, snt.Module]],
        dataset: Iterator[reverb.ReplaySample],
        replay_client: Optional[reverb.Client] = None,
        counter: Optional[counting.Counter] = None,
        logger: Optional[types.NestedLogger] = None,
    ) -> core.Trainer:
        """Creates an instance of the trainer.
        Args:
          networks: struct describing the networks needed by the trainer; this can
            be specific to the trainer in question.
          training_info: Info on what architecture is used. This is necessary for
          the trainer to know how to process
          certain data to work with the specific critics.
          dataset: iterator over samples from replay.
          replay_client: client which allows communication with replay, e.g. in
            order to update priorities.
          counter: a Counter which allows for recording of counts (learner steps,
            actor steps, etc.) distributed throughout the agent.
          logger: Logger object for logging metadata.
        """



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import os
from typing import Any, Dict, Mapping, MutableSequence, Sequence, Union

import numpy as np
import sonnet as snt

# import logging
import tensorflow as tf
import tensorflow_probability as tfp
from acme import types
from acme.tf import networks
from acme.tf import savers as tf2_savers
from acme.tf import utils as tf2_utils

from mava import specs as mava_specs
from mava.components.tf.architectures import StateBasedQValueCritic
from mava.components.tf.networks.mad4pg import DiscreteValuedHead
from mava.utils.environments.RoboCup_env.robocup_utils.util_functions import SpecWrapper

tfd = tfp.distributions


def set_gpu_affinity(gpus: Any) -> Any:
    devices = tf.config.list_physical_devices("GPU")

    # process the requested gpus
    if gpus is None or len(devices) == 0:
        gpus = {}
    elif isinstance(gpus, int):
        gpus = set(devices if gpus == -1 else [devices[gpus]])
    elif isinstance(gpus, MutableSequence):
        gpus = {devices[k] for k in gpus}
    else:
        raise ValueError(
            f"Can't parse requested gpus, got {gpus} of type {type(gpus)}."
        )

    # set the visibility
    # if the set is empty, the function will be called on an empty list
    # so that no `LogicalDevice` will be created on the physical devices.
    tf.config.set_visible_devices(list(gpus), "GPU")
    os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(
        [gpu.name.split(":")[-1] for gpu in gpus]
    )

    # set memory growth
    for gpu in gpus:
        try:
            tf.config.experimental.set_memory_growth(gpu, True)
        except RuntimeError as e:
            print(e)
    return gpus


def make_networks(
    environment_spec: mava_specs.MAEnvironmentSpec,
    policy_networks_layer_sizes: Union[Dict[str, Sequence], Sequence] = (
        256,
        256,
        256,
    ),
    critic_networks_layer_sizes: Union[Dict[str, Sequence], Sequence] = (512, 512, 256),
    shared_weights: bool = True,
    sigma: float = 0.3,
) -> Mapping[str, types.TensorTransformation]:
    """Creates networks used by the agents."""
    specs = environment_spec.get_agent_specs()

    # Create agent_type specs
    if shared_weights:
        type_specs = {key.split("_")[0]: specs[key] for key in specs.keys()}
        specs = type_specs

    if isinstance(policy_networks_layer_sizes, Sequence):
        policy_networks_layer_sizes = {
            key: policy_networks_layer_sizes for key in specs.keys()
        }
    if isinstance(critic_networks_layer_sizes, Sequence):
        critic_networks_layer_sizes = {
            key: critic_networks_layer_sizes for key in specs.keys()
        }

    observation_networks = {}
    policy_networks = {}
    critic_networks = {}
    for key in specs.keys():

        # Get total number of action dimensions from action spec.
        num_dimensions = np.prod(specs[key].actions.shape, dtype=int)

        # Create the shared observation network; here simply a state-less operation.
        observation_network = tf2_utils.to_sonnet_module(tf2_utils.batch_concat)

        # Create the policy network.
        policy_network = snt.Sequential(
            [
                observation_network,
                networks.LayerNormMLP(
                    policy_networks_layer_sizes[key], activate_final=True
                ),
                networks.NearZeroInitializedLinear(num_dimensions),
                networks.TanhToSpec(specs[key].actions),
                networks.ClippedGaussian(sigma),
                networks.ClipToSpec(specs[key].actions),
            ]
        )

        # Create the critic network.
        critic_network = snt.Sequential(
            [
                # The multiplexer concatenates the observations/actions.
                networks.CriticMultiplexer(),
                networks.LayerNormMLP(
                    critic_networks_layer_sizes[key], activate_final=False
                ),
                snt.Linear(1),
            ]
        )
        observation_networks[key] = observation_network
        policy_networks[key] = policy_network
        critic_networks[key] = critic_network

    return {
        "policies": policy_networks,
        "critics": critic_networks,
        "observations": observation_networks,
    }


def make_recurrent_networks(
    environment_spec: mava_specs.MAEnvironmentSpec,
    policy_networks_layer_sizes: Union[Dict[str, Sequence], Sequence] = (
        256,
        256,
        256,
    ),
    critic_networks_layer_sizes: Union[Dict[str, Sequence], Sequence] = (512, 512, 256),
    shared_weights: bool = True,
    sigma: float = 0.3,
    vmin: float = -150.0,
    vmax: float = 150.0,
    num_atoms: int = 51,
) -> Mapping[str, types.TensorTransformation]:
    """Creates networks used by the agents."""
    specs = environment_spec.get_agent_specs()

    # Create agent_type specs
    if shared_weights:
        type_specs = {key.split("_")[0]: specs[key] for key in specs.keys()}
        specs = type_specs

    if isinstance(policy_networks_layer_sizes, Sequence):
        policy_networks_layer_sizes = {
            key: policy_networks_layer_sizes for key in specs.keys()
        }
    if isinstance(critic_networks_layer_sizes, Sequence):
        critic_networks_layer_sizes = {
            key: critic_networks_layer_sizes for key in specs.keys()
        }

    observation_networks = {}
    policy_networks = {}
    critic_networks = {}
    for key in specs.keys():

        # Get total number of action dimensions from action spec.
        num_dimensions = np.prod(specs[key].actions.shape, dtype=int)

        # Create the observation network.
        observation_network = tf2_utils.to_sonnet_module(tf2_utils.batch_concat)

        # Create the policy network.
        policy_network = snt.DeepRNN(
            [
                observation_network,
                snt.Flatten(),
                snt.nets.MLP(policy_networks_layer_sizes[key]),
                snt.LSTM(25),
                snt.nets.MLP([128]),
                networks.NearZeroInitializedLinear(num_dimensions),
                networks.TanhToSpec(specs[key].actions),
                networks.ClippedGaussian(sigma),
                networks.ClipToSpec(specs[key].actions),
            ]
        )

        # Create the critic network.
        critic_network = snt.Sequential(
            [
                # The multiplexer concatenates the observations/actions.
                networks.CriticMultiplexer(),
                networks.LayerNormMLP(
                    critic_networks_layer_sizes[key], activate_final=False
                ),
                DiscreteValuedHead(vmin, vmax, num_atoms),
            ]
        )

        observation_networks[key] = observation_network
        policy_networks[key] = policy_network
        critic_networks[key] = critic_network

    return {
        "observations": observation_networks,
        "policies": policy_networks,
        "critics": critic_networks,
    }


class NNBot(object):
    def __init__(
        self, agent_type, num_players, checkpoint_subpath: str = "Checkpoints"
    ):
        # Convert action and observation specs.
        spec_wrapper = SpecWrapper(num_players)

        # setup the resources
        set_gpu_affinity(0)

        environment_spec = mava_specs.MAEnvironmentSpec(spec_wrapper)

        # Create networks
        shared_weights = True
        system_networks = make_networks(
            environment_spec=environment_spec, shared_weights=shared_weights
        )

        architecture = StateBasedQValueCritic(
            environment_spec=environment_spec,
            policy_networks=system_networks["policies"],
            critic_networks=system_networks["critics"],
            observation_networks=system_networks["observations"],
            shared_weights=shared_weights,
        )

        # Create the policy_networks
        policy_networks = architecture.create_actor_variables()
        before_sum = policy_networks["policies"][agent_type].variables[1].numpy().sum()
        objects_to_save = {
            "policy": policy_networks["policies"][agent_type],
            "observation": policy_networks["observations"][agent_type],
        }

        checkpointer_dir = os.path.join(checkpoint_subpath, agent_type)
        tf2_savers.Checkpointer(
            time_delta_minutes=1,
            add_uid=False,
            directory=checkpointer_dir,
            objects_to_save=objects_to_save,
            enable_checkpointing=True,
        )
        after_sum = policy_networks["policies"][agent_type].variables[1].numpy().sum()

        assert before_sum != after_sum

        self.policy = policy_networks["policies"][agent_type]
        self.num_players = num_players

    def reset_brain(self):
        pass

    def get_action(self, observation):
        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        batched_observation = tf2_utils.add_batch_dim(observation)

        # Compute the policy, conditioned on the observation.
        policy = self.policy(batched_observation)

        # Sample from the policy if it is stochastic.
        action = policy.sample() if isinstance(policy, tfd.Distribution) else policy
        return action.numpy()[0]



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO (Kevin): complete class for targeted communication

"""Broadcasted communication for multi-agent RL systems"""

from typing import Dict

import sonnet as snt

from mava.components.tf.architectures import BaseArchitecture
from mava.components.tf.modules.communication import BaseCommunicationModule


class TargetedCommunication(BaseCommunicationModule):
    """Multi-agent broadcasted communication architecture."""

    def __init__(self, architecture: BaseArchitecture) -> None:
        self._architecture = architecture

    def process_messages(
        self,
        messages: Dict[str, snt.Module],
    ) -> Dict[str, snt.Module]:
        """Perform some communication logic"""



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MADDPG system implementation."""
import copy
import dataclasses
from typing import Any, Dict, Iterator, List, Optional, Type, Union

import reverb
import sonnet as snt
from acme import datasets
from acme.specs import EnvironmentSpec
from acme.tf import variable_utils
from acme.utils import counting, loggers
from dm_env import specs as dm_specs

from mava import adders, core, specs, types
from mava.adders import reverb as reverb_adders
from mava.systems.builders import SystemBuilder
from mava.systems.tf import executors
from mava.systems.tf.maddpg import training
from mava.systems.tf.maddpg.execution import MADDPGFeedForwardExecutor
from mava.wrappers import DetailedTrainerStatistics, NetworkStatisticsActorCritic

BoundedArray = dm_specs.BoundedArray
DiscreteArray = dm_specs.DiscreteArray


@dataclasses.dataclass
class MADDPGConfig:
    """Configuration options for the MADDPG system.
    Args:
            environment_spec: description of the actions, observations, etc.
            policy_networks: the online (optimized) policies for each agent in
                the system.
            critic_networks: the online critic for each agent in the system.
            observation_networks: dictionary of optional networks to transform
                the observations before they are fed into any network.
            discount: discount to use for TD updates.
            batch_size: batch size for updates.
            prefetch_size: size to prefetch from replay.
            target_update_period: number of learner steps to perform before updating
              the target networks.
            min_replay_size: minimum replay size before updating.
            max_replay_size: maximum replay size.
            samples_per_insert: number of samples to take from replay for every insert
              that is made.
            n_step: number of steps to squash into a single transition.
            sigma: standard deviation of zero-mean, Gaussian exploration noise.
            clipping: whether to clip gradients by global norm.
            logger: logger object to be used by trainers.
            counter: counter object used to keep track of steps.
            checkpoint: boolean indicating whether to checkpoint the trainers.
            replay_table_name: string indicating what name to give the replay table."""

    environment_spec: specs.MAEnvironmentSpec
    policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]]
    critic_optimizer: snt.Optimizer
    shared_weights: bool = True
    discount: float = 0.99
    batch_size: int = 256
    prefetch_size: int = 4
    target_averaging: bool = False
    target_update_period: int = 100
    target_update_rate: Optional[float] = None
    executor_variable_update_period: int = 1000
    min_replay_size: int = 1000
    max_replay_size: int = 1000000
    samples_per_insert: float = 32.0
    n_step: int = 5
    sequence_length: int = 20
    period: int = 20
    max_gradient_norm: Optional[float] = None
    sigma: float = 0.3
    logger: loggers.Logger = None
    counter: counting.Counter = None
    checkpoint: bool = True
    checkpoint_subpath: str = "~/mava/"
    replay_table_name: str = reverb_adders.DEFAULT_PRIORITY_TABLE


class MADDPGBuilder(SystemBuilder):
    """Builder for MADDPG which constructs individual components of the system."""

    """Defines an interface for defining the components of an RL system.
      Implementations of this interface contain a complete specification of a
      concrete RL system. An instance of this class can be used to build an
      RL system which interacts with the environment either locally or in a
      distributed setup.
      """

    def __init__(
        self,
        config: MADDPGConfig,
        trainer_fn: Union[
            Type[training.MADDPGBaseTrainer],
            Type[training.MADDPGBaseRecurrentTrainer],
        ] = training.MADDPGDecentralisedTrainer,
        executor_fn: Type[core.Executor] = MADDPGFeedForwardExecutor,
        extra_specs: Dict[str, Any] = {},
    ):
        """Args:
        config: Configuration options for the MADDPG system.
        trainer_fn: Trainer module to use."""

        self._config = config
        self._extra_specs = extra_specs

        """ _agents: a list of the agent specs (ids).
            _agent_types: a list of the types of agents to be used."""
        self._agents = self._config.environment_spec.get_agent_ids()
        self._agent_types = self._config.environment_spec.get_agent_types()
        self._trainer_fn = trainer_fn
        self._executor_fn = executor_fn

    def convert_discrete_to_bounded(
        self, environment_spec: specs.MAEnvironmentSpec
    ) -> specs.MAEnvironmentSpec:
        env_adder_spec: specs.MAEnvironmentSpec = copy.deepcopy(environment_spec)
        keys = env_adder_spec._keys
        for key in keys:
            agent_spec = env_adder_spec._specs[key]
            if type(agent_spec.actions) == DiscreteArray:
                num_actions = agent_spec.actions.num_values
                minimum = [float("-inf")] * num_actions
                maximum = [float("inf")] * num_actions
                new_act_spec = BoundedArray(
                    shape=(num_actions,),
                    minimum=minimum,
                    maximum=maximum,
                    dtype="float32",
                    name="actions",
                )

                env_adder_spec._specs[key] = EnvironmentSpec(
                    observations=agent_spec.observations,
                    actions=new_act_spec,
                    rewards=agent_spec.rewards,
                    discounts=agent_spec.discounts,
                )
        return env_adder_spec

    def make_replay_tables(
        self,
        environment_spec: specs.MAEnvironmentSpec,
    ) -> List[reverb.Table]:
        """Create tables to insert data into."""

        env_adder_spec = self.convert_discrete_to_bounded(environment_spec)

        # Select adder
        if issubclass(self._executor_fn, executors.FeedForwardExecutor):
            adder_sig = reverb_adders.ParallelNStepTransitionAdder.signature(
                env_adder_spec, self._extra_specs
            )
        elif issubclass(self._executor_fn, executors.RecurrentExecutor):
            adder_sig = reverb_adders.ParallelSequenceAdder.signature(
                env_adder_spec, self._extra_specs
            )
        else:
            raise NotImplementedError("Unknown executor type: ", self._executor_fn)

        if self._config.samples_per_insert is None:
            # We will take a samples_per_insert ratio of None to mean that there is
            # no limit, i.e. this only implies a min size limit.
            limiter = reverb.rate_limiters.MinSize(self._config.min_replay_size)

        else:
            # Create enough of an error buffer to give a 10% tolerance in rate.
            samples_per_insert_tolerance = 0.1 * self._config.samples_per_insert
            error_buffer = self._config.min_replay_size * samples_per_insert_tolerance
            limiter = reverb.rate_limiters.SampleToInsertRatio(
                min_size_to_sample=self._config.min_replay_size,
                samples_per_insert=self._config.samples_per_insert,
                error_buffer=error_buffer,
            )

        replay_table = reverb.Table(
            name=self._config.replay_table_name,
            sampler=reverb.selectors.Uniform(),
            remover=reverb.selectors.Fifo(),
            max_size=self._config.max_replay_size,
            rate_limiter=limiter,
            signature=adder_sig,
        )

        return [replay_table]

    def make_dataset_iterator(
        self,
        replay_client: reverb.Client,
    ) -> Iterator[reverb.ReplaySample]:

        sequence_length = (
            self._config.sequence_length
            if issubclass(self._executor_fn, executors.RecurrentExecutor)
            else None
        )

        """Create a dataset iterator to use for learning/updating the system."""
        dataset = datasets.make_reverb_dataset(
            table=self._config.replay_table_name,
            server_address=replay_client.server_address,
            batch_size=self._config.batch_size,
            prefetch_size=self._config.prefetch_size,
            sequence_length=sequence_length,
        )
        return iter(dataset)

    def make_adder(
        self,
        replay_client: reverb.Client,
    ) -> Optional[adders.ParallelAdder]:
        """Create an adder which records data generated by the executor/environment.
        Args:
          replay_client: Reverb Client which points to the replay server.
        """

        # Select adder

        if issubclass(self._executor_fn, executors.FeedForwardExecutor):
            adder = reverb_adders.ParallelNStepTransitionAdder(
                priority_fns=None,
                client=replay_client,
                n_step=self._config.n_step,
                discount=self._config.discount,
            )
        elif issubclass(self._executor_fn, executors.RecurrentExecutor):
            adder = reverb_adders.ParallelSequenceAdder(
                priority_fns=None,
                client=replay_client,
                sequence_length=self._config.sequence_length,
                period=self._config.period,
            )
        else:
            raise NotImplementedError("Unknown executor type: ", self._executor_fn)
        return adder

    def make_executor(
        self,
        policy_networks: Dict[str, snt.Module],
        adder: Optional[adders.ParallelAdder] = None,
        variable_source: Optional[core.VariableSource] = None,
    ) -> core.Executor:
        """Create an executor instance.
        Args:
          policy_networks: A struct of instance of all the different policy networks;
           this should be a callable
            which takes as input observations and returns actions.
          adder: How data is recorded (e.g. added to replay).
          variable_source: A source providing the necessary executor parameters.
        """

        shared_weights = self._config.shared_weights

        variable_client = None
        if variable_source:
            agent_keys = self._agent_types if shared_weights else self._agents

            # Create policy variables
            variables = {}
            for agent in agent_keys:
                variables[agent] = policy_networks[agent].variables

            # Get new policy variables
            variable_client = variable_utils.VariableClient(
                client=variable_source,
                variables={"policy": variables},
                update_period=self._config.executor_variable_update_period,
            )

            # Make sure not to use a random policy after checkpoint restoration by
            # assigning variables before running the environment loop.
            variable_client.update_and_wait()

        # Create the actor which defines how we take actions.
        return self._executor_fn(
            policy_networks=policy_networks,
            agent_specs=self._config.environment_spec.get_agent_specs(),
            shared_weights=shared_weights,
            variable_client=variable_client,
            adder=adder,
        )

    def make_trainer(
        self,
        networks: Dict[str, Dict[str, snt.Module]],
        dataset: Iterator[reverb.ReplaySample],
        replay_client: Optional[reverb.Client] = None,
        counter: Optional[counting.Counter] = None,
        logger: Optional[types.NestedLogger] = None,
        connection_spec: Dict[str, List[str]] = None,
    ) -> core.Trainer:
        """Creates an instance of the trainer.
        Args:
          networks: struct describing the networks needed by the trainer; this can
            be specific to the trainer in question.
          dataset: iterator over samples from replay.
          replay_client: client which allows communication with replay, e.g. in
            order to update priorities.
          counter: a Counter which allows for recording of counts (trainer steps,
            executor steps, etc.) distributed throughout the system.
          logger: Logger object for logging metadata.
        """
        agents = self._agents
        agent_types = self._agent_types
        shared_weights = self._config.shared_weights
        max_gradient_norm = self._config.max_gradient_norm
        discount = self._config.discount
        target_update_period = self._config.target_update_period
        target_averaging = self._config.target_averaging
        target_update_rate = self._config.target_update_rate

        # trainer args
        trainer_config: Dict[str, Any] = {
            "agents": agents,
            "agent_types": agent_types,
            "policy_networks": networks["policies"],
            "critic_networks": networks["critics"],
            "observation_networks": networks["observations"],
            "target_policy_networks": networks["target_policies"],
            "target_critic_networks": networks["target_critics"],
            "target_observation_networks": networks["target_observations"],
            "shared_weights": shared_weights,
            "policy_optimizer": self._config.policy_optimizer,
            "critic_optimizer": self._config.critic_optimizer,
            "max_gradient_norm": max_gradient_norm,
            "discount": discount,
            "target_averaging": target_averaging,
            "target_update_period": target_update_period,
            "target_update_rate": target_update_rate,
            "dataset": dataset,
            "counter": counter,
            "logger": logger,
            "checkpoint": self._config.checkpoint,
            "checkpoint_subpath": self._config.checkpoint_subpath,
        }
        if connection_spec:
            trainer_config["connection_spec"] = connection_spec

        # The learner updates the parameters (and initializes them).
        trainer = self._trainer_fn(**trainer_config)

        # NB If using both NetworkStatistics and TrainerStatistics, order is important.
        # NetworkStatistics needs to appear before TrainerStatistics.
        # TODO(Kale-ab/Arnu): need to fix wrapper type issues
        trainer = NetworkStatisticsActorCritic(trainer)  # type: ignore

        trainer = DetailedTrainerStatistics(  # type: ignore
            trainer, metrics=["policy_loss", "critic_loss"]
        )

        return trainer



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Install script for setuptools."""

import datetime
import sys
from importlib import util as import_util

from setuptools import find_packages, setup

spec = import_util.spec_from_file_location("_metadata", "mava/_metadata.py")
_metadata = import_util.module_from_spec(spec)  # type: ignore
spec.loader.exec_module(_metadata)  # type: ignore

reverb_requirements = [
    "dm-reverb>=0.3.0",
    "jax",
    "jaxlib",
]

tf_requirements = [
    "tensorflow>=2.5.0",
    "tensorflow_probability",
    "dm-sonnet",
    "trfl",
]

env_requirements = [
    "pettingzoo>=1.8.2",
    "multi_agent_ale_py",
    "supersuit",
    "pygame",
    "pysc2",
]

launchpad_requirements = [
    "dm-launchpad-nightly",
]

testing_formatting_requirements = [
    "pre-commit",
    "mypy==0.812",
    "pytest-xdist",
    "flake8==3.9.1",
    "black==21.4b1",
    "pytest-cov",
]

record_episode_requirements = ["array2gif"]

flatland_requirements = ["flatland-rl"]

long_description = """Mava is a library for building multi-agent reinforcement
learning (MARL) systems. Mava builds off of Acme and in a similar way strives
to expose simple, efficient, and readable components, as well as examples that
serve both as reference implementations of popular algorithms and as strong
baselines, while still providing enough flexibility to do novel research.
For more information see
[github repository](https://github.com/instadeepai/mava)."""

# Get the version from metadata.
version = _metadata.__version__  # type: ignore

# If we're releasing a nightly/dev version append to the version string.
if "--nightly" in sys.argv:
    sys.argv.remove("--nightly")
    version += ".dev" + datetime.datetime.now().strftime("%Y%m%d")

setup(
    name="id-mava",
    version=version,
    description="A Python library for Multi-Agent Reinforcement Learning.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="InstaDeep Ltd",
    license="Apache License, Version 2.0",
    keywords="multi-agent reinforcement-learning python machine learning",
    packages=find_packages(),
    install_requires=[
        "dm-acme==0.2.0",
        "absl-py",
        "dm_env",
        "dm-tree",
        "numpy",
        "pillow",
        "matplotlib",
        "dataclasses",
        "Box2D",
    ],
    extras_require={
        "tf": tf_requirements,
        "envs": env_requirements,
        "flatland": flatland_requirements,
        "reverb": reverb_requirements,
        "launchpad": launchpad_requirements,
        "testing_formatting": testing_formatting_requirements,
        "record_episode": record_episode_requirements,
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Dict, List, Tuple, Union

import gym
import numpy as np
from gym import spaces

"""
DIAL switch game implementation
Adapted from https://arxiv.org/pdf/1605.06676.pdf.

Actions:
0 - Nothing
1 - Tell
"""


class MultiAgentSwitchGame(gym.Env):
    def __init__(
        self,
        num_agents: int = 3,
    ) -> None:

        self.num_agents = num_agents
        # Generate IDs and convert agent list to dictionary format.
        self.env_done = False
        self.agent_ids = []

        # spaces
        self.action_spaces = {}
        self.observation_spaces = {}
        for a_i in range(self.num_agents):
            agent_id = "agent_" + str(a_i)
            self.agent_ids.append(agent_id)
            self.action_spaces[agent_id] = spaces.Discrete(2)
            # self.observation_spaces[agent_id] = spaces.Discrete(2)
            self.observation_spaces[agent_id] = spaces.Box(
                low=-np.inf, high=np.inf, shape=(2,), dtype=np.float32
            )

        self.possible_agents = self.agent_ids
        # environment parameters
        self.max_time = 4 * self.num_agents - 6
        self.selected_agent = -1

    def step(
        self, action_n: Dict[str, int]
    ) -> Tuple[
        Dict[str, Union[np.array, Any]],
        Union[dict, Dict[str, Union[float, Any]]],
        Dict[str, Any],
        Dict[str, dict],
    ]:
        obs_n = {}
        reward_n = {}
        done_n = {}

        # set action for interrogated agent
        selected_agent_id = self.agent_ids[self.selected_agent]

        # advance world state
        self.agent_history.append(self.selected_agent)
        self.seen_all = np.unique(self.agent_history).shape[0] == self.num_agents
        if action_n[selected_agent_id] == 1:
            self.env_done = True
            self.tell = True

        self.time += 1
        # self.selected_agent = np.random.randint(0, self.num_agents)

        if self.time >= self.max_time:
            self.env_done = True
        else:
            self.selected_agent = self._agent_order[self.time]

        # record observation for each agent
        for a_i, agent_id in enumerate(self.agent_ids):
            obs_n[agent_id] = self._get_obs(a_i, agent_id)
            reward_n[agent_id] = self._get_reward(a_i, agent_id)
            done_n[agent_id] = self._get_done(agent_id)
            # info_n["n"][agent_id] = self._get_info(agent_id)

            if done_n[agent_id]:
                self.env_done = True

        return obs_n, reward_n, done_n, {}

    def reset(self) -> Dict[str, np.array]:
        # reset world
        self.agent_history: List[int] = []
        self.n_seen = 0
        self.time = 0
        self.tell = False

        self._agent_order = np.array([0])
        while len(np.unique(self._agent_order)) < self.num_agents:
            self._agent_order = np.random.randint(0, self.num_agents, (self.max_time,))

        self.selected_agent = self._agent_order[0]
        # self.selected_agent = np.random.randint(0, self.num_agents)

        self.env_done = False
        # record observations for each agent
        obs_n = {}
        for a_i, agent_id in enumerate(self.agent_ids):
            obs_n[agent_id] = self._get_obs(a_i, agent_id)
        return obs_n

    # get info used for benchmarking
    def _get_info(self, agent_id: str) -> Dict:
        return {}

    # get observation for a particular agent
    def _get_obs(self, a_i: int, agent_id: str) -> np.array:
        selected = 1.0 if a_i == self.selected_agent else 0.0
        return np.array([selected, a_i], dtype=np.float32)

    # get dones for a particular agent
    # unused right now -- agents are allowed to go beyond the viewing screen
    def _get_done(self, agent_id: str) -> bool:
        return self.env_done

    # get reward for a particular agent
    def _get_reward(self, a_i: int, agent_id: str) -> float:
        if not self.env_done:
            return np.array(0.0, dtype=np.float32)
        return (
            np.array(1.0, dtype=np.float32)
            if self.seen_all and self.tell
            else np.array(-1.0, dtype=np.float32)
        )



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO (Kevin): complete class for broadcasted communication

"""Broadcasted communication for multi-agent RL systems"""

from typing import Dict

import sonnet as snt
import tensorflow as tf

from mava.components.tf.architectures import DecentralisedPolicyActor
from mava.components.tf.modules.communication import BaseCommunicationModule


class BroadcastedCommunication(BaseCommunicationModule):
    """Multi-agent broadcasted communication architecture."""

    def __init__(
        self,
        architecture: DecentralisedPolicyActor,
        shared: bool = True,
        channel_size: int = 4,
        channel_noise: float = 0.0,
    ) -> None:
        """Initializes the broadcaster communicator.
        Args:
            architecture: the BaseArchitecture used.
            shared: if a shared communication channel is used.
            channel_noise: stddev of normal noise in channel.
        """
        self._architecture = architecture
        self._shared = shared
        self._channel_noise = channel_noise

    # def create_communication_variables(self) -> Dict[str, Dict[str, snt.Module]]:
    #     print(self._architecture)
    #     pass

    def create_system(
        self,
    ) -> Dict[str, Dict[str, snt.Module]]:
        """Create system architecture with communication by modifying architecture."""
        # networks = self.create_communication_variables()

        return self._architecture.create_system()

    def create_behaviour_policy(self) -> Dict[str, snt.Module]:
        # Note (dries): Can't use the base architecture
        # because it assumes there is an observation network.
        return self._architecture._policy_networks

    def process_messages(
        self,
        messages: Dict[str, snt.Module],
    ) -> Dict[str, snt.Module]:
        """Initializes the broadcaster communicator.
        Args:
            messages: Dict of agent messages.
        """
        if self._shared:
            # Sum of all messages
            channel = tf.math.reduce_sum(tf.nest.flatten(messages), axis=0)
            # Add channel noise if applicable
            channel += tf.random.normal(
                channel.shape, mean=0.0, stddev=self._channel_noise
            )
        else:
            # Concat all messages (might need to retain ordering of dict)
            channel = tf.concat(tf.nest.flatten(messages), axis=0)
            # Add channel noise if applicable
            channel += tf.random.normal(
                channel.shape, mean=0.0, stddev=self._channel_noise
            )

        return {key: channel for key in messages}



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import numpy as np

from mava.utils.environments.RoboCup_env.robocup_utils.util_functions import (
    SpecWrapper,
    deg_rot_to_xy,
)


class NaiveBot(object):
    def __init__(self, agent_type, num_players):
        # Convert action and observation specs.

        spec_wrapper = SpecWrapper(num_players)
        self.scaling = spec_wrapper.scaling
        self.num_players = num_players
        assert self.scaling == 200.0
        self.kicked = False

    def reset_brain(self):
        self.kicked = False

    def get_action(self, observation):
        obs_dict = {
            "time_left": 0,
            "side": 1,
            "sense_self": 2,
            "coords": (3, 5),
            "body_dir": (5, 7),
            "head_dir": (7, 9),
            "width": (9, 12),
            "quality": 12,
            "stamina": 13,
            "effort": 14,
            "speed_amount": 15,
            "speed_direction": (16, 18),
            "neck_direction": (18, 20),
            "see_ball": 20,
            "ball_dist": 21,
            "ball_dir": (22, 24),
            "ball_dist_change": 24,
            "ball_dir_change": 25,
            "ball_speed": 26,
        }
        s, e = obs_dict["coords"]
        p_loc = observation[s:e] * self.scaling

        s, e = obs_dict["body_dir"]
        p_deg = np.arctan2(observation[e - 1], observation[s]) * 180 / np.pi

        # see_ball, ball_distance, ball_direction (x, y format),
        see_ball = observation[obs_dict["see_ball"]]
        b_dist = observation[obs_dict["ball_dist"]] * self.scaling
        s, e = obs_dict["ball_dir"]
        b_deg = np.arctan2(observation[e - 1], observation[s]) * 180 / np.pi
        # Compute the policy, conditioned on the observation.
        # dash (speed), turn (direction in x,y format),
        # kick (direction in x,y format, power)
        command = {
            "do_dash": 0,
            "do_kick": 1,
            "do_change_view": 2,
            "do_tackle": 3,
            "do_turn": 4,
            "do_turn_neck": 5,
            "do_none": 6,
            "dash_power": 7,
            "dash_dir_x": 8,
            "dash_dir_y": 9,
            "turn_dir_x": 18,
            "turn_dir_y": 19,
            "kick_power": 10,
            "kick_dir_x": 11,
            "kick_dir_y": 12,
        }
        action = np.zeros(22)
        if see_ball < 0.5:
            # Search for the ball
            action[command["do_turn"]] = 1
            x, y = deg_rot_to_xy(20)
            action[command["turn_dir_x"]] = x
            action[command["turn_dir_y"]] = y
        elif b_dist < 0.5:
            if not self.kicked:
                # Kick the ball
                self.kicked = True

                goal_locs = np.array([52.5, 0])

                x_diff, y_diff = goal_locs - p_loc

                kick_abs_dir = -np.arctan2(y_diff, x_diff) * 180 / np.pi

                kick_dir = p_deg - 90 + kick_abs_dir

                action[command["do_kick"]] = 1
                action[command["kick_power"]] = 50 / self.scaling

                x, y = deg_rot_to_xy(kick_dir)
                action[command["kick_dir_x"]] = x
                action[command["kick_dir_y"]] = y
            else:
                action[command["do_none"]] = 1
                # print("Waiting..")
        elif np.abs(b_deg) > 10:
            # Turn towards the ball
            action[command["do_turn"]] = 1
            x, y = deg_rot_to_xy(b_deg / 3)
            action[command["turn_dir_x"]] = x
            action[command["turn_dir_y"]] = y
        else:
            # Dash towards the ball
            action[command["do_dash"]] = 1
            speed = 100

            if b_dist < 5:
                speed = 20

            action[command["dash_power"]] = speed / self.scaling
            x, y = deg_rot_to_xy(0)
            action[command["dash_dir_x"]] = x
            action[command["dash_dir_y"]] = y

        if b_dist > 5.0:
            self.kicked = False

        return action



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, Tuple, Union

import numpy as np
from flatland.envs.rail_env import RailEnv

from mava.wrappers.flatland import FlatlandEnvWrapper


def load_flatland_env(env_config: Dict[str, Any]) -> RailEnv:
    """Loads a flatland environment given a config dict. Also, the possible agents in the
    environment are set"""

    env = RailEnv(**env_config)
    env.possible_agents = env.agents[:]

    return env


def flatland_env_factory(
    evaluation: bool = False,
    env_config: Dict[str, Any] = {},
    preprocessor: Callable[
        [Any], Union[np.ndarray, Tuple[np.ndarray], Dict[str, np.ndarray]]
    ] = None,
    include_agent_info: bool = False,
) -> FlatlandEnvWrapper:
    """Loads a flatand environment and wraps it using the flatland wrapper"""

    del evaluation  # since it has same behaviour for both train and eval

    env = load_flatland_env(env_config)
    wrapped_env = FlatlandEnvWrapper(env, preprocessor, include_agent_info)

    return wrapped_env



# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# based on the following:
# https://github.com/deepmind/acme/blob/master/acme/tf/networks/masked_epsilon_greedy.py

"""Adaptation of trfl epsilon_greedy with legal action masking."""
from typing import Optional

import tensorflow as tf
import tensorflow.compat.v1 as tfv1
import tensorflow_probability as tfp
from acme import types as acme_types


def epsilon_greedy_action_selector(
    action_values: acme_types.NestedArray,
    epsilon: Optional[tf.Tensor] = None,
    legal_actions_mask: acme_types.NestedArray = None,
) -> tfp.distributions.Categorical:
    """Computes an epsilon-greedy distribution over actions.
    This returns a categorical distribution over a discrete action space. It is
    assumed that the trailing dimension of `action_values` is of length A, i.e.
    the number of actions. It is also assumed that actions are 0-indexed.
    This policy does the following:
    - With probability 1 - epsilon, take the action corresponding to the highest
    action value, breaking ties uniformly at random.
    - With probability epsilon, take an action uniformly at random.
    Args:
      action_values: A Tensor of action values with any rank >= 1 and dtype float.
        Shape can be flat ([A]), batched ([B, A]), a batch of sequences
        ([T, B, A]), and so on.
      epsilon: A scalar Tensor (or Python float) with value between 0 and 1.
      legal_actions_mask: An optional one-hot tensor having the shame shape and
        dtypes as `action_values`, defining the legal actions:
        legal_actions_mask[..., a] = 1 if a is legal, 0 otherwise.
        If not provided, all actions will be considered legal and
        `tf.ones_like(action_values)`.
    Returns:
      policy: tfp.distributions.Categorical distribution representing the policy.
    """
    with tfv1.name_scope("epsilon_greedy", values=[action_values, epsilon]):

        # Convert inputs to Tensors if they aren't already.
        action_values = tfv1.convert_to_tensor(action_values)

        if epsilon is not None:
            epsilon = tfv1.convert_to_tensor(epsilon, dtype=action_values.dtype)
        else:
            epsilon = 0

        # convert mask to float
        legal_actions_mask = tfv1.cast(legal_actions_mask, dtype=tf.float32)

        # We compute the action space dynamically.
        num_actions = tfv1.cast(tfv1.shape(action_values)[-1], action_values.dtype)

        # Dithering action distribution.
        if legal_actions_mask is None:
            dither_probs = 1 / num_actions * tfv1.ones_like(action_values)
        else:
            dither_probs = (
                1
                / tfv1.reduce_sum(legal_actions_mask, axis=-1, keepdims=True)
                * legal_actions_mask
            )

        # Greedy action distribution, breaking ties uniformly at random.
        max_value = tfv1.reduce_max(action_values, axis=-1, keepdims=True)
        greedy_probs = tfv1.cast(
            tfv1.equal(action_values, max_value), action_values.dtype
        )
        greedy_probs /= (
            tfv1.reduce_sum(greedy_probs, axis=-1, keepdims=True) * legal_actions_mask
        )

        # Epsilon-greedy action distribution.
        probs = epsilon * dither_probs + (1 - epsilon) * greedy_probs

        # Make the policy object.
        policy = tfp.distributions.Categorical(probs=probs)

    return tf.cast(policy.sample(), "int64")



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Adapted from
# https://github.com/deepmind/acme/blob/master/acme/adders/reverb/sequence.py

"""Sequence adders.
This implements adders which add sequences or partial trajectories.
"""

from typing import NamedTuple, Optional

import reverb
import tensorflow as tf
import tree
from acme import specs, types
from acme.adders.reverb import utils
from acme.utils import tree_utils

from mava.adders.reverb import base
from mava.adders.reverb import utils as mava_utils


class StateSpecs(NamedTuple):
    """Container for (observation, legal_actions, terminal) tuples."""

    hidden: types.Nest
    cell: types.Nest


class ParallelSequenceAdder(base.ReverbParallelAdder):
    """An adder which adds sequences of fixed length."""

    def __init__(
        self,
        client: reverb.Client,
        sequence_length: int,
        period: int,
        delta_encoded: bool = False,
        chunk_length: Optional[int] = None,
        priority_fns: Optional[base.PriorityFnMapping] = None,
        pad_end_of_episode: bool = True,
        break_end_of_episode: bool = True,
        max_in_flight_items: Optional[int] = 25,
        use_next_extras: bool = True,
    ):
        """Makes a SequenceAdder instance.
        Args:
          client: See docstring for BaseAdder.
          sequence_length: The fixed length of sequences we wish to add.
          period: The period with which we add sequences. If less than
            sequence_length, overlapping sequences are added. If equal to
            sequence_length, sequences are exactly non-overlapping.
          delta_encoded: If `True` (False by default) enables delta encoding, see
            `Client` for more information.
          chunk_length: Number of timesteps grouped together before delta encoding
            and compression. See `Client` for more information.
          priority_fns: See docstring for BaseAdder.
          pad_end_of_episode: If True (default) then upon end of episode the current
            sequence will be padded (with observations, actions, etc... whose values
            are 0) until its length is `sequence_length`. If False then the last
            sequence in the episode may have length less than `sequence_length`.
          break_end_of_episode: If 'False' (True by default) does not break
            sequences on env reset. In this case 'pad_end_of_episode' is not used.
          max_in_flight_items: The maximum number of items allowed to be "in flight"
            at the same time. See `reverb.Writer.writer` for more info.
        """

        if delta_encoded:
            NotImplementedError(
                "Note (dries): Delta encoding has not been verified to "
                "work in Mava yet. If you want to use delta encoding"
                " first verify that it is working and then contribute "
                "the update with a working example of delta encoding"
                " to the Mava repo."
            )

        super().__init__(
            client=client,
            buffer_size=sequence_length,
            max_sequence_length=sequence_length,
            delta_encoded=delta_encoded,
            chunk_length=chunk_length,
            priority_fns=priority_fns,
            max_in_flight_items=max_in_flight_items,
            use_next_extras=use_next_extras,
        )

        if pad_end_of_episode and not break_end_of_episode:
            raise ValueError(
                "Can't set pad_end_of_episode=True and break_end_of_episode=False at"
                " the same time, since those behaviors are incompatible."
            )
        self._period = period
        self._step = 0
        self._pad_end_of_episode = pad_end_of_episode
        self._break_end_of_episode = break_end_of_episode

    def reset(self) -> None:
        # If we do not break on end of episode, we should not reset the _step
        # counter, neither clear the buffer/writer.
        if self._break_end_of_episode:
            self._step = 0
            super().reset()

    def _write(self) -> None:
        # Append the previous step and increment number of steps written.
        self._writer.append(self._buffer[-1])
        self._step += 1
        self._maybe_add_priorities()

    def _write_last(self) -> None:
        # Create a final step.
        # TODO (Dries): Should self._next_observation be used
        #  here? Should this function be used for sequential?

        final_step = mava_utils.final_step_like(
            self._buffer[0],
            self._next_observations,
            self._next_extras if self._use_next_extras else None,
        )

        # Append the final step.
        self._buffer.append(final_step)
        self._writer.append(final_step)
        self._step += 1

        if not self._break_end_of_episode:
            # Write priorities for the sequence.
            self._maybe_add_priorities()

            # base.py has a check that on add_first self._next_observation should be
            # None, thus we need to clear it at the end of each episode.
            self._next_observations = None
            self._next_extras = None
            return None

        # Determine the delta to the next time we would write a sequence.
        first_write = self._step <= self._max_sequence_length
        if first_write:
            delta = self._max_sequence_length - self._step
        else:
            delta = (
                self._period - (self._step - self._max_sequence_length)
            ) % self._period

        # Bump up to the position where we will write a sequence.
        self._step += delta

        if self._pad_end_of_episode:
            zero_step = tree.map_structure(utils.zeros_like, final_step)

            # Pad with zeros to get a full sequence.
            for _ in range(delta):
                self._buffer.append(zero_step)
                self._writer.append(zero_step)
        elif not first_write:
            # Pop items from the buffer to get a truncated sequence.
            # Note: this is consistent with the padding loop above, since adding zero
            # steps pops the left-most elements. Here we just pop without padding.
            for _ in range(delta):
                self._buffer.popleft()

        # Write priorities for the sequence.
        self._maybe_add_priorities()

    def _maybe_add_priorities(self) -> None:
        if not (
            # Write the first time we hit the max sequence length...
            self._step == self._max_sequence_length
            or
            # ... or every `period`th time after hitting max length.
            (
                self._step > self._max_sequence_length
                and (self._step - self._max_sequence_length) % self._period == 0
            )
        ):
            return

        # Compute priorities for the buffer.
        steps = list(self._buffer)
        num_steps = len(steps)
        table_priorities = utils.calculate_priorities(self._priority_fns, steps)

        # Create a prioritized item for each table.
        for table_name, priority in table_priorities.items():
            self._writer.create_item(table_name, num_steps, priority)

    @classmethod
    def signature(
        cls,
        environment_spec: specs.EnvironmentSpec,
        extras_spec: tf.TypeSpec = {},
    ) -> tf.TypeSpec:
        """This is a helper method for generating signatures for Reverb tables.
        Signatures are useful for validating data types and shapes, see Reverb's
        documentation for details on how they are used.
        Args:
          environment_spec: A `specs.EnvironmentSpec` whose fields are nested
            structures with leaf nodes that have `.shape` and `.dtype` attributes.
            This should come from the environment that will be used to generate
            the data inserted into the Reverb table.
          extras_spec: A nested structure with leaf nodes that have `.shape` and
            `.dtype` attributes. The structure (and shapes/dtypes) of this must
            be the same as the `extras` passed into `ReverbAdder.add`.
        Returns:
          A `Step` whose leaf nodes are `tf.TensorSpec` objects.
        """
        agent_specs = environment_spec.get_agent_specs()
        agents = environment_spec.get_agent_ids()
        env_extras_spec = environment_spec.get_extra_specs()
        extras_spec.update(env_extras_spec)

        obs_specs = {}
        act_specs = {}
        reward_specs = {}
        step_discount_specs = {}
        for agent in agents:
            rewards_spec, step_discounts_spec = tree_utils.broadcast_structures(
                agent_specs[agent].rewards, agent_specs[agent].discounts
            )
            obs_specs[agent] = agent_specs[agent].observations
            act_specs[agent] = agent_specs[agent].actions
            reward_specs[agent] = rewards_spec
            step_discount_specs[agent] = step_discounts_spec

        spec_step = base.Step(
            observations=obs_specs,
            actions=act_specs,
            rewards=reward_specs,
            discounts=step_discount_specs,
            start_of_episode=specs.Array(shape=(), dtype=bool),
            extras=extras_spec,
        )
        return tree.map_structure_with_path(base.spec_like_to_tensor_spec, spec_step)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common types used throughout Mava."""

from typing import Dict, NamedTuple, Union

import numpy as np
from acme import types
from acme.utils import loggers


class OLT(NamedTuple):
    """Container for (observation, legal_actions, terminal) tuples."""

    observation: types.Nest
    legal_actions: types.Nest
    terminal: types.Nest


NestedLogger = Union[loggers.Logger, Dict[str, loggers.Logger]]

SingleAgentAction = Union[int, float, np.array]
Action = Union[SingleAgentAction, Dict[str, SingleAgentAction]]

SingleAgentReward = Union[int, float]
Reward = Union[SingleAgentReward, Dict[str, SingleAgentReward]]
Discount = Reward

Observation = Union[OLT, Dict[str, OLT]]



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qmix trainer implementation."""
import time
from typing import Any, Dict, List, Optional, Sequence

import numpy as np
import sonnet as snt
import tensorflow as tf
from acme.tf import utils as tf2_utils
from acme.utils import counting, loggers
from trfl.indexing_ops import batched_index

from mava.components.tf.modules.communication import BaseCommunicationModule
from mava.components.tf.modules.exploration.exploration_scheduling import (
    LinearExplorationScheduler,
)

# from mava.systems.tf import savers as tf2_savers
from mava.systems.tf.madqn.training import MADQNTrainer
from mava.utils import training_utils as train_utils

train_utils.set_growing_gpu_memory()


class QMIXTrainer(MADQNTrainer):
    """QMIX trainer.
    This is the trainer component of a QMIX system. i.e. it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        q_networks: Dict[str, snt.Module],
        target_q_networks: Dict[str, snt.Module],
        mixing_network: snt.Module,
        target_mixing_network: snt.Module,
        target_update_period: int,
        dataset: tf.data.Dataset,
        optimizer: snt.Optimizer,
        discount: float,
        shared_weights: bool,
        exploration_scheduler: LinearExplorationScheduler,
        communication_module: Optional[BaseCommunicationModule] = None,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        fingerprint: bool = False,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
    ) -> None:

        self._mixing_network = mixing_network
        self._target_mixing_network = target_mixing_network
        self._optimizer = optimizer

        super(QMIXTrainer, self).__init__(
            agents=agents,
            agent_types=agent_types,
            q_networks=q_networks,
            target_q_networks=target_q_networks,
            target_update_period=target_update_period,
            dataset=dataset,
            optimizer=optimizer,
            discount=discount,
            shared_weights=shared_weights,
            exploration_scheduler=exploration_scheduler,
            communication_module=communication_module,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            fingerprint=fingerprint,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
        )

        # Checkpoint the mixing networks
        # TODO add checkpointing for mixing networks

    def _update_target_networks(self) -> None:
        # Update target networks (incl. mixing networks).
        if tf.math.mod(self._num_steps, self._target_update_period) == 0:
            for key in self.unique_net_keys:
                online_variables = [
                    *self._q_networks[key].variables,
                ]
                target_variables = [
                    *self._target_q_networks[key].variables,
                ]

                # Make online -> target network update ops.
                for src, dest in zip(online_variables, target_variables):
                    dest.assign(src)

            # NOTE These shouldn't really be in the agent for loop.
            online_variables = [
                *self._mixing_network.variables,
                # *self._mixing_network._hypernetworks.variables,
            ]
            target_variables = [
                *self._target_mixing_network.variables,
                # *self._target_mixing_network._hypernetworks.variables,
            ]
            # Make online -> target network update ops.
            for src, dest in zip(online_variables, target_variables):
                dest.assign(src)

        self._num_steps.assign_add(1)

    def step(self) -> None:
        # Run the learning step.
        fetches = self._step()

        # Compute elapsed time.
        timestamp = time.time()
        if self._timestamp:
            elapsed_time = timestamp - self._timestamp
        else:
            elapsed_time = 0
        self._timestamp = timestamp  # type: ignore

        # Update our counts and record it.
        counts = self._counter.increment(steps=1, walltime=elapsed_time)
        fetches.update(counts)

        # Checkpoint and attempt to write the logs.
        if self._checkpoint:
            train_utils.checkpoint_networks(self._system_checkpointer)

        # Log and decrement epsilon
        epsilon = self.get_epsilon()
        fetches["epsilon"] = epsilon
        self._decrement_epsilon()

        if self._logger:
            self._logger.write(fetches)

    @tf.function
    def _step(
        self,
    ) -> Dict[str, Dict[str, Any]]:

        # Update the target networks
        self._update_target_networks()

        # Get data from replay (dropping extras if any). Note there is no
        # extra data here because we do not insert any into Reverb.
        inputs = next(self._iterator)

        self._forward(inputs)

        self._backward()

        # Log losses per agent
        return {agent: {"q_value_loss": self.loss} for agent in self._agents}

    def _forward(self, inputs: Any) -> None:
        # Unpack input data as follows:
        # o_tm1 = dictionary of observations one for each agent
        # a_tm1 = dictionary of actions taken from obs in o_tm1
        # e_tm1 [Optional] = extra data that the agents persist in replay.
        # r_t = dictionary of rewards or rewards sequences
        #   (if using N step transitions) ensuing from actions a_tm1
        # d_t = environment discount ensuing from actions a_tm1.
        #   This discount is applied to future rewards after r_t.
        # o_t = dictionary of next observations or next observation sequences
        # e_t = [Optional] = extra data that the agents persist in replay.
        o_tm1, a_tm1, e_tm1, r_t, d_t, o_t, e_t = inputs.data
        s_tm1 = e_tm1["s_t"]
        s_t = e_t["s_t"]

        # Do forward passes through the networks and calculate the losses
        with tf.GradientTape(persistent=True) as tape:
            q_acts = []  # Q vals
            q_targets = []  # Target Q vals
            for agent in self._agents:
                agent_key = self.agent_net_keys[agent]

                o_tm1_feed, o_t_feed, a_tm1_feed = self._get_feed(
                    o_tm1, o_t, a_tm1, agent
                )
                q_tm1 = self._q_networks[agent_key](o_tm1_feed)
                q_t_value = self._target_q_networks[agent_key](o_t_feed)
                q_t_selector = self._q_networks[agent_key](o_t_feed)
                best_action = tf.argmax(q_t_selector, axis=1, output_type=tf.int32)

                # TODO Make use of q_t_selector for fingerprinting. Speak to Claude.
                q_act = batched_index(q_tm1, a_tm1_feed, keepdims=True)  # [B, 1]
                q_target = batched_index(
                    q_t_value, best_action, keepdims=True
                )  # [B, 1]

                q_acts.append(q_act)
                q_targets.append(q_target)

            rewards = tf.concat(
                [tf.reshape(val, (-1, 1)) for val in list(r_t.values())], axis=1
            )
            rewards = tf.reduce_mean(rewards, axis=1)  # [B]

            pcont = tf.concat(
                [tf.reshape(val, (-1, 1)) for val in list(d_t.values())], axis=1
            )
            pcont = tf.reduce_mean(pcont, axis=1)
            discount = tf.cast(self._discount, list(d_t.values())[0].dtype)
            pcont = discount * pcont  # [B]

            q_acts = tf.concat(q_acts, axis=1)  # [B, num_agents]
            q_targets = tf.concat(q_targets, axis=1)  # [B, num_agents]

            q_tot_mixed = self._mixing_network(q_acts, s_tm1)  # [B, 1, 1]
            q_tot_target_mixed = self._target_mixing_network(
                q_targets, s_t
            )  # [B, 1, 1]

            # Calculate Q loss.
            targets = rewards + pcont * q_tot_target_mixed
            targets = tf.stop_gradient(targets)
            td_error = targets - q_tot_mixed

            # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
            self.loss = 0.5 * tf.reduce_mean(tf.square(td_error))
            self.tape = tape

    def _backward(self) -> None:
        for agent in self._agents:
            agent_key = self.agent_net_keys[agent]

            # Update agent networks
            variables = [*self._q_networks[agent_key].trainable_variables]
            gradients = self.tape.gradient(self.loss, variables)
            gradients = tf.clip_by_global_norm(gradients, self._max_gradient_norm)[0]
            self._optimizers[agent_key].apply(gradients, variables)

        # Update mixing network
        variables = [
            *self._mixing_network.trainable_variables,
            # *self._mixing_network._hypernetworks.trainable_variables,
        ]
        gradients = self.tape.gradient(self.loss, variables)
        gradients = tf.clip_by_global_norm(gradients, self._max_gradient_norm)[0]
        self._optimizer.apply(gradients, variables)

        train_utils.safe_del(self, "tape")

    def get_variables(self, names: Sequence[str]) -> Dict[str, Dict[str, np.ndarray]]:
        variables: Dict[str, Dict[str, np.ndarray]] = {}
        variables = {}
        for network_type in names:
            if network_type == "mixing":
                # Includes the hypernet variables
                variables[network_type] = self._mixing_network.variables
            else:  # Collect variables for each agent network
                variables[network_type] = {
                    key: tf2_utils.to_numpy(
                        self._system_network_variables[network_type][key]
                    )
                    for key in self.unique_net_keys
                }
        return variables



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple, Union

import dm_env
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp
from acme import types
from acme.specs import EnvironmentSpec

# Internal imports.
from acme.tf import utils as tf2_utils
from acme.tf import variable_utils as tf2_variable_utils
from dm_env import specs

from mava import adders
from mava.systems.tf import executors

Array = specs.Array
BoundedArray = specs.BoundedArray
DiscreteArray = specs.DiscreteArray

tfd = tfp.distributions


class MADDPGFeedForwardExecutor(executors.FeedForwardExecutor):
    """A feed-forward executor for discrete actions in MADDPG.
    An executor based on a feed-forward policy for each agent in the system
    which takes non-batched observations and outputs non-batched actions.
    It also allows adding experiences to replay and updating the weights
    from the policy on the learner.
    """

    def __init__(
        self,
        policy_networks: Dict[str, snt.Module],
        agent_specs: Dict[str, EnvironmentSpec],
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        shared_weights: bool = True,
    ):

        """Initializes the executor.
        Args:
          networks: the (recurrent) policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
        """

        # Store these for later use.
        self._agent_specs = agent_specs
        super().__init__(
            policy_networks=policy_networks,
            shared_weights=shared_weights,
            adder=adder,
            variable_client=variable_client,
        )

    @tf.function
    def _policy(
        self, agent: str, observation: types.NestedTensor
    ) -> types.NestedTensor:

        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        batched_observation = tf2_utils.add_batch_dim(observation)

        # index network either on agent type or on agent id
        agent_key = agent.split("_")[0] if self._shared_weights else agent

        # Compute the policy, conditioned on the observation.
        policy = self._policy_networks[agent_key](batched_observation)

        # TODO (dries): Make this support hybrid action spaces.
        if type(self._agent_specs[agent].actions) == BoundedArray:
            # Continuous action
            action = policy
        elif type(self._agent_specs[agent].actions) == DiscreteArray:
            action = tf.math.argmax(policy, axis=1)
        else:
            raise NotImplementedError

        return action, policy

    def select_action(
        self, agent: str, observation: types.NestedArray
    ) -> Tuple[types.NestedArray, types.NestedArray]:

        # Step the recurrent policy/value network forward
        # given the current observation and state.
        action, policy = self._policy(agent, observation.observation)

        # Return a numpy array with squeezed out batch dimension.
        action = tf2_utils.to_numpy_squeeze(action)
        policy = tf2_utils.to_numpy_squeeze(policy)
        return action, policy

    def select_actions(
        self, observations: Dict[str, types.NestedArray]
    ) -> Tuple[Dict[str, types.NestedArray], Dict[str, types.NestedArray]]:

        actions = {}
        policies = {}
        for agent, observation in observations.items():
            actions[agent], policies[agent] = self.select_action(agent, observation)
        return actions, policies

    def observe(
        self,
        actions: Union[
            Dict[str, types.NestedArray], List[Dict[str, types.NestedArray]]
        ],
        next_timestep: dm_env.TimeStep,
        next_extras: Dict[str, types.NestedArray] = {},
    ) -> None:
        if self._adder:
            _, policy = actions
            # TODO (dries): Sort out this mypy issue.
            self._adder.add(policy, next_timestep, next_extras)  # type: ignore


class MADDPGRecurrentExecutor(executors.RecurrentExecutor):
    def __init__(
        self,
        policy_networks: Dict[str, snt.Module],
        agent_specs: Dict[str, EnvironmentSpec],
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        shared_weights: bool = True,
        store_recurrent_state: bool = True,
    ):

        """Initializes the executor.
        Args:
          networks: the (recurrent) policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
        """

        # Store these for later use.
        self._agent_specs = agent_specs

        super().__init__(
            policy_networks=policy_networks,
            shared_weights=shared_weights,
            adder=adder,
            variable_client=variable_client,
            store_recurrent_state=store_recurrent_state,
        )

    @tf.function
    def _policy(
        self,
        agent: str,
        observation: types.NestedTensor,
        state: types.NestedTensor,
    ) -> Tuple[types.NestedTensor, types.NestedTensor, types.NestedTensor]:

        # Add a dummy batch dimension and as a side effect convert numpy to TF.
        batched_observation = tf2_utils.add_batch_dim(observation)

        # index network either on agent type or on agent id
        agent_key = agent.split("_")[0] if self._shared_weights else agent

        # Compute the policy, conditioned on the observation.
        policy, new_state = self._policy_networks[agent_key](batched_observation, state)

        # TODO (dries): Make this support hybrid action spaces.
        if type(self._agent_specs[agent].actions) == BoundedArray:
            # Continuous action
            action = policy
        elif type(self._agent_specs[agent].actions) == DiscreteArray:
            action = tf.math.argmax(policy, axis=1)
        else:
            raise NotImplementedError

        return action, policy, new_state

    def select_action(
        self, agent: str, observation: types.NestedArray
    ) -> types.NestedArray:

        # TODO Mask actions here using observation.legal_actions
        # Initialize the RNN state if necessary.
        if self._states[agent] is None:
            # index network either on agent type or on agent id
            agent_key = agent.split("_")[0] if self._shared_weights else agent
            self._states[agent] = self._policy_networks[agent_key].initia_state(1)

        # Step the recurrent policy forward given the current observation and state.
        action, policy, new_state = self._policy(
            agent, observation.observation, self._states[agent]
        )

        # Bookkeeping of recurrent states for the observe method.
        self._update_state(agent, new_state)

        # Return a numpy array with squeezed out batch dimension.
        action = tf2_utils.to_numpy_squeeze(action)
        policy = tf2_utils.to_numpy_squeeze(policy)
        return action, policy

    def select_actions(
        self, observations: Dict[str, types.NestedArray]
    ) -> Tuple[Dict[str, types.NestedArray], Dict[str, types.NestedArray]]:

        actions = {}
        policies = {}
        for agent, observation in observations.items():
            actions[agent], policies[agent] = self.select_action(agent, observation)
        return actions, policies

    def observe(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        next_extras: Optional[Dict[str, types.NestedArray]] = {},
    ) -> None:
        if not self._adder:
            return
        _, policy = actions
        if not self._store_recurrent_state:
            if next_extras:
                # TODO (dries): Sort out this mypy issue.
                self._adder.add(policy, next_timestep, next_extras)  # type: ignore
            else:
                self._adder.add(policy, next_timestep)  # type: ignore
            return

        numpy_states = {
            agent: tf2_utils.to_numpy_squeeze(_state)
            for agent, _state in self._states.items()
        }
        if next_extras:
            next_extras.update({"core_states": numpy_states})
            self._adder.add(policy, next_timestep, next_extras)  # type: ignore
        else:
            self._adder.add(policy, next_timestep, numpy_states)  # type: ignore
